//////////////////////////////////////////////////////////////////////////////
// RGG LED main.c
//////////////////////////////////////////////////////////////////////////////


#include "compiler.h"
#include "preprocessor.h"
#include "pm.h"
#include "gpio.h"
#include "pwm.h"
#include "adc.h"

//#include "print_funcs.h"
#include "intc.h"
#include "power_clocks_lib.h"
#include "conf_usb.h"
#include "usb_task.h"

#if USB_DEVICE_FEATURE == ENABLED
#include "usb_drv.h"
#include "usb_descriptors.h"
#include "usb_standard_request.h"
#include "device_task.h"
#endif

//////////////////////////////////////////////////////////////////////////////
// clock setup
//////////////////////////////////////////////////////////////////////////////

#define FOSC0           12000000                              //!< Osc0 frequency: Hz.
#define OSC0_STARTUP    AVR32_PM_OSCCTRL0_STARTUP_2048_RCOSC  //!< Osc0 startup time: RCOsc periods.
void init_clock() {
	// source: gpio_local_bus_example.c

	// Initialize domain clocks (CPU, HSB, PBA and PBB) to the max frequency available
	// without flash wait states.
	// Some of the registers in the GPIO module are mapped onto the CPU local bus.
	// To ensure maximum transfer speed and cycle determinism, any slaves being
	// addressed by the CPU on the local bus must be able to receive and transmit
	// data on the bus at CPU clock speeds. The consequences of this is that the
	// GPIO module has to run at the CPU clock frequency when local bus transfers
	// are being performed => we want fPBA = fCPU.

	// Switch the main clock source to Osc0.
	pm_switch_to_osc0(&AVR32_PM, FOSC0, OSC0_STARTUP);

	// Setup PLL0 on Osc0, mul=10 ,no divisor, lockcount=16: 12Mhzx11 = 132MHz output
	pm_pll_setup(&AVR32_PM, 0, // pll.
			10, // mul.
			1, // div.
			0, // osc.
			16); // lockcount.
	// PLL output VCO frequency is 132MHz.
	// We divide it by 2 with the pll_div2=1 to get a main clock at 66MHz.
	pm_pll_set_option(&AVR32_PM, 0, // pll.
			1, // pll_freq.
			1, // pll_div2.
			0); // pll_wbwdisable.
	// Enable the PLL.
	pm_pll_enable(&AVR32_PM, 0);
	// Wait until the PLL output is stable.
	pm_wait_for_pll0_locked(&AVR32_PM);
	// Configure each clock domain to use the main clock divided by 2
	// => fCPU = fPBA = fPBB = 33MHz.
	pm_cksel(&AVR32_PM, 1, // pbadiv.
			0, // pbasel.
			1, // pbbdiv.
			0, // pbbsel.
			1, // hsbdiv=cpudiv
			0); // hsbsel=cpusel
	// Switch the main clock source to PLL0.
	pm_switch_to_clock(&AVR32_PM, AVR32_PM_MCCTRL_MCSEL_PLL0);
}

//////////////////////////////////////////////////////////////////////////////
// PWM setup
//////////////////////////////////////////////////////////////////////////////

#define cR 					0
#define cG 					1
#define cB 					2

#define R_PWM_PIN			AVR32_PWM_1_0_PIN
#define R_PWM_FUNCTION		AVR32_PWM_1_0_FUNCTION
#define R_PWM_CHANNEL_ID	1

#define G_PWM_PIN			AVR32_PWM_3_0_PIN
#define G_PWM_FUNCTION		AVR32_PWM_3_0_FUNCTION
#define G_PWM_CHANNEL_ID	3

#define B_PWM_PIN			AVR32_PWM_4_0_PIN
#define B_PWM_FUNCTION		AVR32_PWM_4_0_FUNCTION
#define B_PWM_CHANNEL_ID	4

// storage:
static pwm_opt_t pwm_opt;
static avr32_pwm_channel_t pwm_channel[3];
static unsigned int channel_id[3];
// WARNING! the PWM example app does:
//avr32_pwm_channel_t pwm_channel = { .ccnt = 0 };  // One channel config.
// we don't, as pwm_channel_init() does not use the .ccnt field...!?!?

#define TTT 4

void init_pwm() {
	// set PWM GPIOs
	gpio_enable_module_pin(R_PWM_PIN, R_PWM_FUNCTION);
	gpio_enable_module_pin(G_PWM_PIN, G_PWM_FUNCTION);
	gpio_enable_module_pin(B_PWM_PIN, B_PWM_FUNCTION);

	// PWM controller configuration.
	pwm_opt.diva = AVR32_PWM_DIVA_CLK_OFF;
	pwm_opt.divb = AVR32_PWM_DIVB_CLK_OFF;
	pwm_opt.prea = AVR32_PWM_PREA_MCK;
	pwm_opt.preb = AVR32_PWM_PREB_MCK;
	// init pwm globals
	pwm_init(&pwm_opt);

	channel_id[cR] = R_PWM_CHANNEL_ID;
	channel_id[cG] = G_PWM_CHANNEL_ID;
	channel_id[cB] = B_PWM_CHANNEL_ID;

	unsigned int c;
	for (c = cR; c <= cB; c++) {

		/*
		 pwm_channel[c].CMR.calg = PWM_MODE_LEFT_ALIGNED; // Channel mode.
		 pwm_channel[c].CMR.cpol = PWM_POLARITY_LOW; // Channel polarity.
		 pwm_channel[c].CMR.cpd = PWM_UPDATE_DUTY; // Not used the first time.
		 pwm_channel[c].CMR.cpre = AVR32_PWM_CPRE_MCK_DIV_256; // Channel prescaler.
		 pwm_channel[c].cdty = 5; // Channel duty cycle, should be < CPRD.
		 pwm_channel[c].cprd = 20; // Channel period.
		 pwm_channel[c].cupd = 0; // Channel update is not used here.
		 // With these settings, the output waveform period will be :
		 // (115200/256)/20 == 22.5Hz == (MCK/prescaler)/period, with MCK == 115200Hz,
		 // prescaler == 256, period == 20.
		 */

		pwm_channel[c].CMR.calg = PWM_MODE_LEFT_ALIGNED; // Channel mode.
		pwm_channel[c].CMR.cpol = PWM_POLARITY_HIGH; // Channel polarity.
		pwm_channel[c].CMR.cpd = PWM_UPDATE_DUTY; // Not used the first time.
		pwm_channel[c].CMR.cpre = AVR32_PWM_CPRE_MCK_DIV_2; // Channel prescaler.
		pwm_channel[c].cdty = 0; // Channel duty cycle, should be < CPRD.
		pwm_channel[c].cprd = (256 << TTT); // Channel period.
		pwm_channel[c].cupd = 0; // Channel update is not used here.

		pwm_channel_init(channel_id[c], &pwm_channel[c]);
	}

	pwm_start_channels((1 << channel_id[cR]) | (1 << channel_id[cG]) | (1
			<< channel_id[cB]));
}

void set_rgb(U8 r, U8 g, U8 b) {
	U8 c;
	for (c = cR; c <= cB; c++) {
		// Channel duty cycle, should be < CPRD.
		switch (c) {
		case cR:
			pwm_channel[c].cdty = ((U32) r) << TTT;
			break;
		case cG:
			pwm_channel[c].cdty = ((U32) g) << TTT;
			break;
		case cB:
			pwm_channel[c].cdty = ((U32) b) << TTT;
			break;
		}
		pwm_channel_init(channel_id[c], &pwm_channel[c]);
	}
}

//////////////////////////////////////////////////////////////////////////////
// ADC
//////////////////////////////////////////////////////////////////////////////

#define ADC_CHANNEL     0
#define ADC_PIN         AVR32_ADC_AD_0_PIN
#define ADC_FUNCTION    AVR32_ADC_AD_0_FUNCTION

void init_adc() {
	// GPIO pin/adc-function map.
	static const gpio_map_t ADC_GPIO_MAP = { { ADC_PIN, ADC_FUNCTION } };

	volatile avr32_adc_t *adc = &AVR32_ADC; // ADC IP registers address


	// Assign the on-board sensors to their ADC channel.
	unsigned short adc_channel = ADC_CHANNEL;

	//		clock already set up
	//	  // switch to oscillator 0
	//	  pm_switch_to_osc0(&AVR32_PM, FOSC0, OSC0_STARTUP);

	//		we are not using serial debugging
	//	  // init debug serial line
	//	  init_dbg_rs232(FOSC0);

	// Assign and enable GPIO pins to the ADC function.
	gpio_enable_module(ADC_GPIO_MAP, sizeof(ADC_GPIO_MAP)
			/ sizeof(ADC_GPIO_MAP[0]));

	// configure ADC
	// Lower the ADC clock to match the ADC characteristics (because we configured
	// the CPU clock to 33MHz, and the ADC clock requires less than 5 MHz for 10 bit ADC,
	// therefore prescale by ;
	// cf. the ADC Characteristic section in the datasheet).
	// These numbers are wrong currently!
	AVR32_ADC.mr |= 0x3 << AVR32_ADC_MR_PRESCAL_OFFSET;
	adc_configure(adc);

	// Enable the ADC channels.
	adc_enable(adc, adc_channel);
}

U16 get_adc_value() {
    // launch conversion on all enabled channels
	volatile avr32_adc_t *adc = &AVR32_ADC; // ADC IP registers address
    adc_start(adc);

    // get value for the adc channel
    U16 adc_value = adc_get_value(adc, ADC_CHANNEL);

    return adc_value;
}

//////////////////////////////////////////////////////////////////////////////
// USB
//////////////////////////////////////////////////////////////////////////////

static U32 sof_cnt;

static U8 in_data_length;
static U8 in_buf[EP_SIZE_TEMP1];

static U8 out_data_length;
static U8 out_buf[EP_SIZE_TEMP2];

//!
//! @brief This function initializes the hardware/software resources required for device applicative task.
//!
void device_task_init(void) {
	sof_cnt = 0;
	in_data_length = 0;
	out_data_length = 0;

	Usb_enable_sof_interrupt();
}

//!
//! @brief Entry point of the device applicative task management
//!
//! This function links the device application to the USB bus.
//!

void device_task(void) {
	// First, check the device enumeration state
	if (!Is_device_enumerated())
		return;

	// HERE STARTS THE USB DEVICE APPLICATIVE CODE
	// The example below just performs a loopback transmission/reception.
	// All data received with the OUT endpoint is stored in a RAM buffer and
	// sent back to the IN endpoint.


	// Load the IN endpoint with the contents of the RAM buffer
	if (Is_usb_in_ready(EP_TEMP_IN)) {
		gpio_local_tgl_gpio_pin(AVR32_PIN_PA11);

		// read ADC and store to buffer...:
		U16 adcval = get_adc_value();
		in_buf[0] = 0xDE;
		in_buf[1] = 0xAD;
		in_buf[2] = 0xFF & (adcval >> 8);
		in_buf[3] = 0xFF & (adcval >> 0);
		in_data_length = 4;

		Usb_reset_endpoint_fifo_access(EP_TEMP_IN);
		usb_write_ep_txpacket(EP_TEMP_IN, in_buf, in_data_length, NULL);
		in_data_length = 0;
		Usb_ack_in_ready_send(EP_TEMP_IN);

	}

	// If we receive something in the OUT endpoint, just store it in the RAM buffer
	if (Is_usb_out_received(EP_TEMP_OUT)) {
		gpio_local_tgl_gpio_pin(AVR32_PIN_PA12);

		Usb_reset_endpoint_fifo_access(EP_TEMP_OUT);
		out_data_length = Usb_byte_count(EP_TEMP_OUT);
		usb_read_ep_rxpacket(EP_TEMP_OUT, out_buf, out_data_length, NULL);
		Usb_ack_out_received_free(EP_TEMP_OUT);

		// update PWM:
		set_rgb(out_buf[1], out_buf[2], out_buf[3]);
	}

}

//!
//! @brief usb_sof_action
//!
//! This function increments the sof_cnt counter each time
//! the USB Start-of-Frame interrupt subroutine is executed (1 ms).
//! Useful to manage time delays
//!
void usb_sof_action(void) {
	gpio_local_tgl_gpio_pin(AVR32_PIN_PA10);
	sof_cnt++;
}

//////////////////////////////////////////////////////////////////////////////
// main
//////////////////////////////////////////////////////////////////////////////

/*
 unsigned int xx() {
 unsigned int r = rand();
 r = r % 256;
 r = r * r / 256;
 r = r * r / 256;
 return r;
 }
 */

int main() {

	Enable_global_exception();
	INTC_init_interrupts();

	pcl_switch_to_osc(PCL_OSC0, FOSC0, OSC0_STARTUP);

	init_clock();
	init_pwm();
	gpio_local_init();

	init_adc();

	pcl_configure_usb_clock();

	usb_task_init();

#if USB_DEVICE_FEATURE == ENABLED
	device_task_init();
#endif

	gpio_local_enable_pin_output_driver(AVR32_PIN_PA10);
	gpio_local_enable_pin_output_driver(AVR32_PIN_PA11);
	gpio_local_enable_pin_output_driver(AVR32_PIN_PA12);

	gpio_local_clr_gpio_pin(AVR32_PIN_PA10);
	gpio_local_clr_gpio_pin(AVR32_PIN_PA11);
	gpio_local_clr_gpio_pin(AVR32_PIN_PA12);

	while (TRUE) {
		usb_task();

#if USB_DEVICE_FEATURE == ENABLED
		device_task();
#endif
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
// EOF
//////////////////////////////////////////////////////////////////////////////
